%{
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>
	#include "symbol_table.h"
	#include "tree.h"
	#include "code_gen.h"
%}

%start		Program

%token		NUMBER ID
%token		FUNC END GOTO IF RETURN AND NOT

@attributes 	{ struct symbol_t *defined_labels; struct symbol_t *referenced_labels; } Stats Funcdef
@attributes	{ struct symbol_t *defined_labels; } Labels
@attributes	{ struct symbol_t *referenced_labels; treenode *node; } Stat
@attributes	{ char *name; } ID NUMBER
@attributes	{ struct symbol_t *variables; } Ids
@attributes	{ treenode *node; } Expr Term Plusterm Malterm Andterm Array

@traversal @postorder t
@traversal @preorder u
@traversal @postorder v
%%

Program:	  Funcdef ';' Program
       		|
       		;

Funcdef:	  FUNC ID '{' Ids '}' Stats END
       		@{ @i @Funcdef.defined_labels@ = clone_table(@Stats.defined_labels@);
		   @i @Funcdef.referenced_labels@ = clone_table(@Stats.referenced_labels@);
		
		   @t if(!contains_all(@Stats.defined_labels@, @Stats.referenced_labels@)) exit(3);
		   @t if(contains_duplicates(@Stats.defined_labels@)) exit(3);
		   @u function_header(@ID.name@); func_par_index=0; clear_register_assignments();
		   @v function_footer();
		@}
       		;

Ids:		  ID Ids
   		@{ @i @Ids.0.variables@ = table_add_symbol(clone_table(@Ids.1.variables@), @ID.name@);

		   @t if(check_symbol_defined(@Ids.1.variables@, @ID.name@)) exit(3);
		   @u function_parameter(@ID.name@);
		@}

   		|
		@{ @i @Ids.variables@ = new_table();
		@}

   		;

Stats:		  Labels Stat ';' Stats
     		@{ @i @Stats.0.defined_labels@ = table_merge(@Labels.defined_labels@, @Stats.1.defined_labels@);
		   @i @Stats.0.referenced_labels@ = table_merge(@Stat.referenced_labels@, @Stats.1.referenced_labels@);

		   @u burm_label(@Stat.node@); burm_reduce(@Stat.node@, 1);
		@}

     		|
		@{ @i @Stats.defined_labels@ = new_table();
		   @i @Stats.referenced_labels@ = new_table();
		@}

		;

Labels:		  Labels ID ':'
		@{ @i @Labels.0.defined_labels@ = table_add_symbol(@Labels.1.defined_labels@, @ID.name@);
		@}

      		| 
		@{ @i @Labels.defined_labels@ = new_table();
		@}

		;

Stat:		  Lexpr '=' Expr
    		@{ @i @Stat.referenced_labels@ = new_table();
		   @i @Stat.node@ = (treenode *)NULL;
		@}

    		| GOTO ID
		@{ @i @Stat.referenced_labels@ = table_add_symbol(new_table(), @ID.name@);
		   @i @Stat.node@ = (treenode *)NULL;
		@}

		| IF Expr GOTO ID
		@{ @i @Stat.referenced_labels@ = table_add_symbol(new_table(), @ID.name@);
		   @i @Stat.node@ = (treenode *)NULL;
		@}

		| Call
    		@{ @i @Stat.referenced_labels@ = new_table();
		   @i @Stat.node@ = (treenode *)NULL;
		@}
		
		| RETURN Expr
    		@{ @i @Stat.referenced_labels@ = new_table();
                   @i @Stat.node@=new_node(OP_Return, @Expr.node@, (treenode *)NULL);
		@}
		
		;

Lexpr:		  ID
     		| Array
		;

Expr:		  Term
    		@{ @i @Expr.node@ = @Term.node@;
		@}
		
		| NOT Term
		@{ @i @Expr.node@ = new_node(OP_Not, @Term.node@, (treenode *)NULL);
		@}

		| Term '-' Term
		@{ @i @Expr.node@ = new_node(OP_Difference, @Term.0.node@, @Term.1.node@);
		@}

		| Term Plusterm
		@{ @i @Expr.node@ = new_node(OP_Addition, @Term.node@, @Plusterm.node@);
		@}

		| Term Malterm
		@{ @i @Expr.node@ = new_node(OP_Multiplication, @Term.node@, @Malterm.node@);
		@}
		
		| Term Andterm
		@{ @i @Expr.node@ = new_node(OP_Conjunction, @Term.node@, @Andterm.node@);
		@}
		
		| Term '<' Term
		@{ @i @Expr.node@ = new_node(OP_Smaller, @Term.0.node@, @Term.1.node@);
		@}
		
		| Term '=' Term
		@{ @i @Expr.node@ = new_node(OP_Equal, @Term.0.node@, @Term.1.node@);
		@}

		;

Plusterm:	  '+' Term Plusterm
		@{ @i @Plusterm.node@ = new_node(OP_Addition, @Term.node@, @Plusterm.1.node@);
		@}
		
		| '+' Term
		@{ @i @Plusterm.node@ = @Term.node@;
		@}

		;

Malterm:	  '*' Term Malterm
       		@{ @i @Malterm.0.node@ = new_node(OP_Multiplication, @Term.node@, @Malterm.1.node@);
		@}
		
       		| '*' Term
		@{ @i @Malterm.node@ = @Term.node@;
		@}
		
		;

Andterm:	  AND Term Andterm
       		@{ @i @Andterm.0.node@ = new_node(OP_Conjunction, @Term.node@, @Andterm.1.node@);
		@}
		
       		| AND Term
		@{ @i @Andterm.node@ = @Term.node@;
		@}
		
		;

Term:		  '(' Expr ')'
    		@{ @i @Term.node@ = @Expr.node@;
		@}
		
    		| ID
		@{ @i @Term.node@ = new_named_leaf(OP_ID, @ID.name@);
		@}

		| NUMBER
		@{ @i @Term.node@ = new_number_leaf(@NUMBER.name@);
		@}
		
		| Call
		@{ @i @Term.node@ = (treenode *)NULL;
		@}
		
		| Array
		@{ @i @Term.node@ = @Array.node@;
		@}

		;

Array:		  Term '[' Expr ']'
     		@{ @i @Array.node@ = new_node(OP_Array, @Term.node@, @Expr.node@);
		@}

     		;

Call:		  ID '{' Terme '}'
    		;

Terme:		  Term Terme
     		|
		;

%%

extern int yylex();

int yyerror(char *error_text) {
	exit(2);
}


